import logging
import os
import time
from functools import wraps
from typing import Awaitable, Callable, List, Optional

from prometheus_client import Counter, Histogram

logger = logging.getLogger(__name__)


class BasicTimeCostMetrics:
    """
    如果启动prometheus则使用此类
    """

    def __init__(self, buckets: Optional[List[float]] = None):
        default_buckets = [2, 5, 10, 30, 60, 120, 180, 240, 300]
        self.buckets = buckets if buckets is not None else default_buckets
        self.pid = os.getpid()

    def add(self, key, cost): ...

    def add_error(self, key): ...


class ApiTimeCostMetrics(BasicTimeCostMetrics):
    """
    接口操作的耗时统计
    指标名如下：
    operation_duration_seconds_count
    operation_duration_seconds_bucket
    operation_errors_total
    """

    def __init__(self, buckets: Optional[List[float]] = None):
        super().__init__(buckets=buckets)

        self.ops_duration = Histogram(
            "operation_duration_seconds",
            "Api operation duration in seconds",
            ["operation", "pid"],
            buckets=self.buckets,
        )
        self.ops_error_cnt = Counter(
            "operation_errors_total",
            "The total number of errors in operations.",
            ["operation", "pid"],
        )

    def add(self, key, cost):
        self.ops_duration.labels(operation=key, pid=self.pid).observe(cost)

    def add_error(self, key):
        self.ops_error_cnt.labels(operation=key, pid=self.pid).inc(1)


class ClientTimeCostMetrics(BasicTimeCostMetrics):
    """
    统计client请求的耗时
    """

    def __init__(self, buckets: Optional[List[float]] = None):
        super().__init__(buckets=buckets)

        self.duration = Histogram(
            "client_duration_seconds",
            "Client response duration in seconds",
            ["client", "pid"],
            buckets=self.buckets,
        )
        self.error_cnt = Counter(
            "client_errors_total",
            "The total number of errors in client response.",
            ["client", "pid"],
        )

    def add(self, key, cost):
        self.duration.labels(client=key, pid=self.pid).observe(cost)

    def add_error(self, key):
        self.error_cnt.labels(client=key, pid=self.pid).inc(1)


class PrometheusTimeCostMetricSetting:
    __default_metrics = BasicTimeCostMetrics()
    __api_metrics = None
    __client_metrics = None

    @classmethod
    def initialize(
        cls,
        *,
        api_metrics_buckets: Optional[List[int]] = None,
        client_metrics_buckets: Optional[List[int]] = None,
    ):
        if api_metrics_buckets is None:
            cls.__api_metrics = ApiTimeCostMetrics()
        elif len(api_metrics_buckets) == 0:
            cls.__api_metrics = BasicTimeCostMetrics()
        else:
            cls.__api_metrics = ApiTimeCostMetrics(buckets=list(api_metrics_buckets))

        if client_metrics_buckets is None:
            cls.__client_metrics = ClientTimeCostMetrics()
        elif len(client_metrics_buckets) == 0:
            cls.__client_metrics = BasicTimeCostMetrics()
        else:
            cls.__client_metrics = ClientTimeCostMetrics(
                buckets=list(client_metrics_buckets)
            )

    @classmethod
    def api_metrics_instance(cls) -> BasicTimeCostMetrics:
        return cls.__api_metrics or cls.__default_metrics

    @classmethod
    def client_metrics_instance(cls) -> BasicTimeCostMetrics:
        return cls.__client_metrics or cls.__default_metrics

    @classmethod
    def metrics_handler(cls, key: str, metrics: BasicTimeCostMetrics):
        def decorator(func: Callable[..., Awaitable]) -> Callable[..., Awaitable]:
            @wraps(func)
            async def wrapper(*args, **kwargs):
                try:
                    start = time.perf_counter()
                    result = await func(*args, **kwargs)
                    cost = time.perf_counter() - start
                    metrics.add(key, cost)
                    return result
                except Exception as e:
                    metrics.add_error(key)
                    raise e

            return wrapper

        return decorator

    @classmethod
    def client_metrics_handler(cls, client_name: str):
        return cls.metrics_handler(client_name, cls.client_metrics_instance())

    @classmethod
    def api_metrics_handler(cls, api_name: str):
        return cls.metrics_handler(api_name, cls.api_metrics_instance())



import logging
import os
import shutil

from fastapi import FastAPI

logger = logging.getLogger(__name__)


class PrometheusContext:
    PROM_MULTI_PROC_OS_VAR = "PROMETHEUS_MULTIPROC_DIR"  #
    DEFAULT_PROMETHEUS_MULTIPROC_DIR = "/temp/prometheus_multiproc_dir"

    @classmethod
    def get_prometheus_multiproc_dir(cls):
        return os.environ.get(
            cls.PROM_MULTI_PROC_OS_VAR, cls.DEFAULT_PROMETHEUS_MULTIPROC_DIR
        )

    @classmethod
    def init(cls):
        """
        Initialize the Prometheus multiprocessing directory.
        """
        prom_dir = os.environ.get(cls.PROM_MULTI_PROC_OS_VAR, None)
        if not prom_dir:
            logger.warning(
                f"Prometheus multiprocessing directory is not set. Using default directory {cls.DEFAULT_PROMETHEUS_MULTIPROC_DIR}."
            )
            prom_dir = cls.DEFAULT_PROMETHEUS_MULTIPROC_DIR
            os.environ["prometheus_multiproc_dir"] = prom_dir
        else:
            logger.info(
                f"Using Prometheus multiprocessing directory from environment variable: {prom_dir}"
            )

        if os.path.exists(prom_dir):
            shutil.rmtree(prom_dir)
            logger.info(
                f"Removed existing Prometheus multiprocessing directory: {prom_dir}"
            )

        os.makedirs(prom_dir, exist_ok=True)
        logger.info(f"Created new Prometheus multiprocessing directory: {prom_dir}")

    @classmethod
    def mount_prometheus_endpoint(cls, app: FastAPI):
        """
        Mount the Prometheus metrics endpoint on the given application.
        """
        from prometheus_client import CollectorRegistry, make_asgi_app, multiprocess

        registry = CollectorRegistry()
        multiprocess.MultiProcessCollector(registry=registry)
        app.mount("/metrics", make_asgi_app(registry=registry))



import os

from algo_backend.utils import OsAttrMeta


class ErrorCodeConfig:
    SERVICE_PREFIX: int = os.getenv("ERROR_CODE_SERVICE_PREFIX", "0")


class ServiceConfig(metaclass=OsAttrMeta):
    HTTP_PORT: int = 8100
    TIMEOUT_KEEP_ALIVE: int = 1000
    PROCESS_NUM: int = 1



import logging
import time
import traceback
from functools import wraps
from typing import Awaitable, Callable, get_type_hints

from algo_backend.exception import (
    BasicApiId,
    DefaultApiErrorCode,
)
from algo_backend.metrics import PrometheusTimeCostMetricSetting
from algo_backend.schema import AbstractRespVo
from .exception_to_vo import gen_vo_from_exception

logger = logging.getLogger(__name__)


def timing_and_exception_handler(
    func=None,
    *,
    api_id: BasicApiId = DefaultApiErrorCode.DEFAULT_ERROR,
    api_name: str = "",
):
    """
    装饰器：用于统计函数执行时间并捕获异常
    函数中需要包含参数reqid或者request_id
    """

    def decorator(
        func: Callable[..., Awaitable[AbstractRespVo]],
    ) -> Callable[..., Awaitable[AbstractRespVo]]:
        @wraps(func)
        async def wrapper(*args, **kwargs):
            start_time = time.perf_counter()

            request_id = kwargs.get("request_id", None) or kwargs.get("reqid", None)
            _name = func.__name__ or api_name

            try:
                # 执行原函数
                logger.info(f"ReqId: {request_id} | Function: {_name} | Start")
                result: AbstractRespVo = await func(*args, **kwargs)
                # 计算耗时
                elapsed_time = time.perf_counter() - start_time
                logger.info(
                    f"ReqId: {request_id} | Function: {_name} | COST:{elapsed_time:.4f}s"
                )
                PrometheusTimeCostMetricSetting.api_metrics_instance().add(
                    _name, elapsed_time
                )
                return result.set_request_id(request_id)
            except Exception as e:
                # 计算耗时
                elapsed_time = time.perf_counter() - start_time
                PrometheusTimeCostMetricSetting.api_metrics_instance().add_error(_name)
                # 记录异常信息和完整堆栈
                logger.error(
                    f"ReqId: {request_id} | Function: {_name} | COST:{elapsed_time:.4f}s | Exception: {str(e)}\n"
                    f"Traceback:\n{traceback.format_exc()}"
                )

                vo_cls: type(AbstractRespVo) = get_type_hints(func).get(
                    "return"
                )  # 这里可能会失败，因为无法强制用户的类型
                return gen_vo_from_exception(
                    vo_cls, e, api_name=_name, request_id=request_id, api_id=api_id
                )

        return wrapper

    return decorator if func is None else decorator(func)



import logging
from contextlib import asynccontextmanager
from typing import Any, Callable, Coroutine, Dict, List, Optional

from fastapi import FastAPI
from fastapi.exceptions import RequestValidationError, StarletteHTTPException
from starlette.middleware import Middleware

from algo_backend.exception import ApiErrorCodeManage
from algo_backend.intercept import (
    BasicExceptionInterceptor,
    HTTPExceptionInterceptor,
    ValidateExceptionInterceptor,
)
from algo_backend.log import BasicLogStarter
from algo_backend.metrics import (
    PrometheusContext,
    PrometheusTimeCostMetricSetting,
)
from algo_backend.metrics.collector import MetricsScheduleMonitor
from algo_backend.middleware import default_cors_middleware

from .event_list import EventList

logger = logging.getLogger(__name__)


class DefaultAlgoAppGenerator:
    """
    FastApi的App初始化，作用于每个fastapi的worker
    """

    def __init__(
        self,
        *,
        service_name: str,
        app_error_code_prefix: Optional[int] = None,
        custom_start_event: List[Callable[[], Coroutine[Any, Any, None]]] = None,
        custom_end_event: List[Callable[[], Coroutine[Any, Any, None]]] = None,
        log_starter: Optional[BasicLogStarter] = None,
        intercept_dict: Optional[Dict[Exception, BasicExceptionInterceptor]] = None,
        middlewares: List[Middleware] = (default_cors_middleware,),
        api_time_cost_buckets: Optional[List[int]] = None,
        client_time_cost_buckets: Optional[List[int]] = None,
    ):
        """
        :param service_name: 服务名称，必填
        :param app_error_code_prefix: 错误码前缀，两位数字，如果指定则错误码为8位，如果不填则错误码是6位
        :param custom_start_event: 用户自定义的app时的触发事件列表
        :param custom_end_event: 用户自定义的app结束的触发事件列表
        :param log_starter: 日志启动器
        :param intercept_dict: 拦截器字典，key为异常类型，value为拦截器实例
        :param middlewares: 中间件列表
        :param api_time_cost_buckets: 接口耗时统计的桶,None表示使用默认的,空列表表示不启用这个指标
        :param client_time_cost_buckets: 客户端耗时统计的桶,None表示使用默认的,空列表表示不启用这个指标
        """
        # 注册8位错误码前缀
        ApiErrorCodeManage.set_error_code_prefix_env(app_error_code_prefix)
        self.service_name = service_name

        self.__app: Optional[FastAPI] = FastAPI()

        self.start_event_list = EventList()
        self.end_event_list = EventList()

        self.log_starter = log_starter or BasicLogStarter(service_name=service_name)

        self.intercept_dict = {
            RequestValidationError: ValidateExceptionInterceptor(),
            StarletteHTTPException: HTTPExceptionInterceptor(),
        }
        self.intercept_dict.update(intercept_dict or {})

        self.middleware_list = list(middlewares)

        self.time_cost_metrics_initializer = (
            lambda: PrometheusTimeCostMetricSetting.initialize(
                api_metrics_buckets=api_time_cost_buckets,
                client_metrics_buckets=client_time_cost_buckets,
            )
        )

        self.init_start_event(custom_start_event)
        self.init_end_event(custom_end_event)

    def init_start_event(self, custom_event):
        """
        初始化开始事件
        """
        self.start_event_list.add(self.log_starter.setup_log())
        # 启动系统指标、gc指标的定时采集写入以供多进程时的汇总
        self.start_event_list.add(MetricsScheduleMonitor().run_monitor)
        # 接口计时器和客户端解释器的初始化
        self.start_event_list.add(self.time_cost_metrics_initializer)
        # 插入自定义的开始事件
        for event in custom_event or []:
            self.start_event_list.add(event)

    def init_end_event(self, custom_event):
        """
        初始化结束事件
        """
        # 插入自定义事件
        for event in custom_event or []:
            self.end_event_list.add(event)

    def add_middleware(self, middleware: Middleware) -> "DefaultAlgoAppGenerator":
        """
        添加中间件
        """
        self.middleware_list.append(middleware)
        return self

    def set_log_stater(self, log_starter: BasicLogStarter) -> "DefaultAlgoAppGenerator":
        self.log_starter = log_starter
        return self

    def use_log_loguru(self):
        """
        使用loguru日志框架
        """
        from algo_backend.log.loguru import LoguruStarter

        return self.set_log_stater(LoguruStarter(service_name=self.service_name))

    def use_log_nblog(self):
        """
        使用nblog日志框架，待实现
        """
        ...

    def generate(self):
        self.init_app().add_prometheus_endpoint().add_interceptor()
        return self.app

    @property
    def app(self):
        return self.__app

    def init_app(self, **kwargs):
        self.__app = FastAPI(
            lifespan=self.gen_life_span(), middleware=self.middleware_list, **kwargs
        )
        return self

    def add_prometheus_endpoint(self):
        PrometheusContext.mount_prometheus_endpoint(self.__app)
        return self

    def add_interceptor(self):
        for exc, interceptor in self.intercept_dict.items():
            self.__app.add_exception_handler(exc, interceptor.intercept)
        return self

    def gen_life_span(self):
        @asynccontextmanager
        async def lifespan(app: FastAPI):
            logger.info("ExecuteStartEvent")
            await self.start_event_list.execute()
            logger.info("Start App")

            yield

            logger.info("ExecuteStopEvent")
            await self.end_event_list.execute()

            logger.info("Goodbye")

        return lifespan



from dotenv import load_dotenv

load_dotenv(".env")

from .basic_config import ErrorCodeConfig, ServiceConfig
from .loguru_config import LoguruConfig

__all__ = ["LoguruConfig", "ServiceConfig", "ErrorCodeConfig"]



from typing import Optional, List

from algo_backend.metrics import RequestTimeCostMiddleware
from starlette.middleware import Middleware


def http_request_time_cost_middleware(
    buckets: Optional[List[float]] = None, ignore_paths: Optional[List[str]] = None
):
    return Middleware(
        RequestTimeCostMiddleware, buckets=buckets, ignore_paths=ignore_paths
    )



import asyncio
import logging
import os
import threading
from abc import ABC
from typing import List, Type

import psutil
from prometheus_client import Gauge

from .common import AbstractMetricCollector

logger = logging.getLogger(__name__)


class AbstractSystemMetricCollector(AbstractMetricCollector, ABC):
    """系统指标收集器抽象基类"""

    def __init__(self, process: psutil.Process, pid: int, **kwargs):
        super().__init__()
        self.process = process
        self.pid = pid


class CPUMetricCollector(AbstractSystemMetricCollector):
    """CPU使用率指标收集器"""

    def __init__(self, process: psutil.Process, pid: int):
        super().__init__(process, pid)
        self.gauge = Gauge(
            "cpu_usage_percent",
            "CPU usage in percent",
            ["pid"],
            multiprocess_mode="all",
        )

    async def collect(self):
        cpu_usage = self.process.cpu_percent()
        self.gauge.labels(pid=self.pid).set(cpu_usage)


class MemoryMetricCollector(AbstractSystemMetricCollector):
    """内存使用量指标收集器"""

    def __init__(self, process: psutil.Process, pid: int):
        super().__init__(process, pid)
        self.gauge_rss = Gauge(
            "resident_memory_usage_mb",
            "Resident Memory usage in MB",
            ["pid"],
            multiprocess_mode="all",
        )
        self.gauge_vms = Gauge(
            "virtual_memory_usage_mb",
            "Virtual Memory usage in MB",
            ["pid"],
            multiprocess_mode="all",
        )

    async def collect(self):
        memory_info = self.process.memory_info()
        rss = memory_info.rss / (1024 * 1024)  # resident set size resident set size
        vms = memory_info.vms / (1024 * 1024)  # virtual memory size
        self.gauge_rss.labels(pid=self.pid).set(rss)
        self.gauge_vms.labels(pid=self.pid).set(vms)


class ThreadCountMetricCollector(AbstractSystemMetricCollector):
    """线程数指标收集器"""

    def __init__(self, process: psutil.Process, pid: int):
        super().__init__(process, pid)
        self.gauge_active_thread_count = Gauge(
            "active_thread_count",
            "Number of active threads",
            ["pid"],
            multiprocess_mode="all",
        )
        self.gauge_pid_thread_count = Gauge(
            "thread_count", "Number of Pid threads", ["pid"], multiprocess_mode="all"
        )

    async def collect(self):
        pid_thread_count = self.process.num_threads()
        self.gauge_pid_thread_count.labels(pid=self.pid).set(pid_thread_count)
        self.gauge_active_thread_count.labels(pid=self.pid).set(
            threading.active_count()
        )


class FileMetricCollector(AbstractSystemMetricCollector):
    """文件描述符数指标收集器"""

    def __init__(self, process: psutil.Process, pid: int):
        super().__init__(process, pid)
        self.gauge = Gauge(
            "open_fds",
            "Number of open files",
            ["pid"],
            multiprocess_mode="all",
        )

    async def collect(self):
        file_descriptor_count = len(self.process.open_files())
        self.gauge.labels(pid=self.pid).set(file_descriptor_count)


class SystemMetricsMonitor(AbstractMetricCollector):
    """系统指标监控器 - 采用组装模式"""

    def __init__(self, interval_sec: int = 30):
        super().__init__(interval_sec=interval_sec)
        self.pid = os.getpid()
        self.process = psutil.Process(self.pid)
        self.collectors: List[AbstractSystemMetricCollector] = []

    def add(
        self, collector_classes: Type[AbstractSystemMetricCollector], **kwargs
    ) -> "SystemMetricsMonitor":
        self.collectors.append(collector_classes(self.process, self.pid, **kwargs))
        return self

    def cpu(self):
        return self.add(CPUMetricCollector)

    def memory(self):
        return self.add(MemoryMetricCollector)

    def thread(self):
        return self.add(ThreadCountMetricCollector)

    def file(self):
        return self.add(FileMetricCollector)

    def register_default_collectors(self) -> "SystemMetricsMonitor":
        """
        组装默认指标收集器
        """
        self.add(CPUMetricCollector)
        self.add(MemoryMetricCollector)
        self.add(ThreadCountMetricCollector)
        self.add(FileMetricCollector)
        return self

    async def collect(self):
        for collector in self.collectors:
            await collector.collect()

    async def run_monitor(self, interval_sec: int = 10):
        """运行监控任务"""

        self.set_interval(interval_sec)

        async def work():
            while True:
                # 执行所有收集器的收集任务
                for collector in self.collectors:
                    await collector.collect()

                await asyncio.sleep(interval_sec)

        if len(self.collectors) == 0:
            self.register_default_collectors()

        logger.info(
            f"Start system metrics monitor, interval: {interval_sec}s, collector num = {len(self.collectors)}"
        )

        asyncio.create_task(work())



from typing import Any, Callable, List

import uvicorn

from algo_backend.config import ServiceConfig
from algo_backend.metrics import PrometheusContext


class DefaultAlgoServiceStarter:
    """
    主线程的设置
    uvicorn容器启动fastapi
    """

    def __init__(self, service_name: str):
        self.service_name = service_name
        self.__main_pid_event: List[Callable[[], Any]] = []
        self.init_events()

    def init_events(self):
        # 初始化prometheus多进程的设置
        self.append_event(PrometheusContext.init)
        return self

    def append_event(self, func: Callable[[], Any]):
        """尾部插入事件"""
        self.__main_pid_event.append(func)
        return self

    def insert_event(self, func: Callable[[], Any], index: int = 0):
        """任意位置插入事件"""
        self.__main_pid_event.insert(index, func)
        return self

    def use_loguru(self):
        """
        使用loguru日志设置，对主进程的日志生效
        注册loguru的定时清理器
        """
        from algo_backend.log.loguru import LoguruStarter

        loguru_stater = LoguruStarter(service_name=self.service_name)
        self.insert_event(loguru_stater.setup_log)
        self.append_event(loguru_stater.run_log_cleaner)
        return self

    def main_pid_setup(self) -> "DefaultAlgoServiceStarter":
        """
        主进程中的设置事件
        """
        for func in self.__main_pid_event:
            func()
        return self

    @classmethod
    def run(
        cls,
        app_str: str = "app.backend.application:app",
        host: str = "0.0.0.0",
        **kwargs,
    ):
        # 启动服务
        uvicorn.run(
            app_str,
            host=host,
            port=ServiceConfig.HTTP_PORT,
            timeout_keep_alive=ServiceConfig.TIMEOUT_KEEP_ALIVE,
            workers=ServiceConfig.PROCESS_NUM,
            **kwargs,
        )



from .common import AbstractMetricCollector
from .gc_metrics import PythonGcMetricsCollector, PythonGcObjectMetricsCollector
from .schedule_monitor import MetricsScheduleMonitor
from .system_metrics import SystemMetricsMonitor

__all__ = [
    "PythonGcMetricsCollector",
    "PythonGcObjectMetricsCollector",
    "AbstractMetricCollector",
    "MetricsScheduleMonitor",
    "SystemMetricsMonitor",
]



import uuid

from fastapi import Header


def gen_random_request_id():
    return str(uuid.uuid4())


async def get_request_id(
    x_request_id: str = Header(gen_random_request_id(), alias="x-request-id"),
):
    """
    使用示例

    @app.post("/example/help")
    async def run(body: Body, reqid=Depends(get_request_id)):
        ...

    如果请求头中没有x-request-id，则会触发validate_interceptor的检验
    """
    return x_request_id



from abc import ABC, abstractmethod
from typing import Generic, Optional, TypeVar

from pydantic import BaseModel, Field

from algo_backend.exception import BasicException, BasicStatusCode, CommonStatusCode

T = TypeVar("T")


class AbstractRespVo(BaseModel, ABC):
    @classmethod
    @abstractmethod
    def success(cls, result: Optional[T]) -> "AbstractRespVo": ...

    @classmethod
    @abstractmethod
    def from_exception(
        cls,
        e: BasicException,
        *,
        api_name: Optional[str] = None,
        request_id: Optional[str] = None,
    ) -> "AbstractRespVo":
        """
        将异常转换为vo
        """
        ...

    @abstractmethod
    def set_request_id(self, request_id: Optional[str]) -> "AbstractRespVo":
        """
        增加requestId，以帮助追踪问题
        """
        ...


class BaseRespVo(AbstractRespVo, Generic[T]):
    """
    默认vo实现
    """

    code: int
    message: str
    result: Optional[T] = Field(None, description="结果")
    requestId: Optional[str] = None

    @classmethod
    def success(cls, result: Optional[T]) -> "BaseRespVo":
        return BaseRespVo(
            code=CommonStatusCode.SUCCESS.code,
            result=result,
            message=CommonStatusCode.SUCCESS.msg,
        )

    @classmethod
    def from_status_code(cls, status_code: BasicStatusCode, **kwargs) -> "BaseRespVo":
        return BaseRespVo(
            code=status_code.code,
            message=status_code.msg.format(**kwargs),
            result=None,
        )

    @classmethod
    def from_exception(
        cls,
        e: BasicException,
        *,
        api_name: Optional[str] = None,
        request_id: Optional[str] = None,
    ) -> "BaseRespVo":
        vo = BaseRespVo(
            code=e.code,
            message=e.msg,
            result=None,
        )
        vo.set_request_id(request_id)
        vo.message = f"接口{api_name}报错:{vo.message}" if api_name else vo.message
        return vo

    def set_request_id(self, request_id: Optional[str]) -> "BaseRespVo":
        self.requestId = request_id
        return self



import inspect
import sys
from typing import Optional

from algo_backend.config import ErrorCodeConfig

from . import status_code
from .status_code import BasicApiId, BasicApiInnerErrorCode, BasicStatusCode


class ApiErrorCodeManage:
    """
    构建”5aaabb“错误码，aaa接口id，bb表示接口内部错误码
    """

    @classmethod
    def set_error_code_prefix_env(cls, prefix: Optional[int] = None):
        """
        服务启动时执行这个函数
        也可以启动服务时通过环境变量ERROR_CODE_SERVICE_PREFIX指定服务前缀
        如果输入的prefix code不符合规则，则忽略，保持原来的6位
        """
        if prefix and 1 <= prefix <= 99:
            ErrorCodeConfig.SERVICE_PREFIX = prefix

    @classmethod
    def scan_module_and_summary(cls, *list_module) -> list:
        """扫描模块中的ApiId和错误码枚举类"""
        api_id_dict: dict = {}
        api_err_code_info = []

        cls_list = []
        for m in list_module:
            cls_list.extend(
                [obj for name, obj in inspect.getmembers(m) if inspect.isclass(obj)]
            )

        for obj in set(cls_list):
            # 查找继承自BasicApiId的枚举类
            if issubclass(obj, BasicApiId) and obj != BasicApiId:
                for enum_member in obj:
                    api_id_dict[enum_member.value] = enum_member.name
                    api_err_code_info.append(
                        [
                            enum_member.value,
                            BasicApiInnerErrorCode.gen_standard_code(enum_member, 0),
                            "",
                        ]
                    )

            # 查找继承自BasicApiInnerErrorCode的枚举类
            elif (
                issubclass(obj, BasicApiInnerErrorCode)
                and obj != BasicApiInnerErrorCode
                and hasattr(obj, "__api_id__")
            ):
                api_id_enum: BasicApiId = getattr(obj, "__api_id__")
                for enum_member in obj:
                    api_err_code_info.append(
                        [
                            api_id_enum.value,
                            BasicApiInnerErrorCode.gen_standard_code(
                                api_id_enum, enum_member.value
                            ),
                            enum_member.msg,
                        ]
                    )

        for idx, (api_code, _, _) in enumerate(api_err_code_info):
            api_err_code_info[idx][0] = api_id_dict[api_code]

        # 排序
        api_err_code_info.sort(key=lambda x: x[1])

        return api_err_code_info

    @classmethod
    def summary_module_api_error_code_markdown(cls, *list_module):
        """
        打印模块中的ApiId和错误码枚举类
        """
        result = cls.scan_module_and_summary(*list_module)

        print("| ApiName | ErrorCode | Message |")
        print("|--------|------------|---------|")
        for api_id, api_code, msg in result:
            print(f"| {api_id} | {api_code} | {msg} |")


class BasicCodeManage:
    __DEFAULT_MODULE = sys.modules[status_code.__name__]

    @classmethod
    def scan_module_and_summary(cls, *list_module) -> list:
        """扫描模块中BasicStatusCode枚举类及其子类"""
        if cls.__DEFAULT_MODULE not in list_module:
            list_module = list_module + (cls.__DEFAULT_MODULE,)

        result = []
        cls_list = []

        for m in list_module:
            cls_list.extend(
                [obj for name, obj in inspect.getmembers(m) if inspect.isclass(obj)]
            )

        for obj in set(cls_list):
            if issubclass(obj, BasicStatusCode) and obj != BasicApiId:
                for enum_member in obj:
                    result.append([enum_member.value, enum_member.msg])

        result.sort(key=lambda x: x[0])

        return result

    @classmethod
    def summary_error_code_markdown(cls, *list_module):
        """
        打印模块中的ApiId和错误码枚举类
        """
        result = cls.scan_module_and_summary(*list_module)

        print("| Code | Message |")
        print("|--------|------------|")
        for code, msg in result:
            print(f"| {code} | {msg} |")



from .collector import AbstractMetricCollector, SystemMetricsMonitor
from .http_metrics import RequestTimeCostMiddleware
from .prometheus_context import PrometheusContext
from .time_cost_metrics import (
    BasicTimeCostMetrics,
    ApiTimeCostMetrics,
    ClientTimeCostMetrics,
    PrometheusTimeCostMetricSetting,
)

__all__ = [
    "SystemMetricsMonitor",
    "RequestTimeCostMiddleware",
    "PrometheusContext",
    # 定时收集指标
    "AbstractMetricCollector",
    # 耗时指标
    "PrometheusTimeCostMetricSetting",
    "BasicTimeCostMetrics",
    "ApiTimeCostMetrics",
    "ClientTimeCostMetrics",
]



import asyncio
from typing import Any, Callable, Coroutine, List, Optional, Union


class EventList:
    def __init__(self):
        """
        普通函数和协程函数列表
        """
        self.__events: List[
            Union[Callable, Callable[[], Coroutine[Any, Any, None]]]
        ] = []

    def add(
        self,
        event: Optional[
            Union[Callable, Callable[[], Coroutine[Any, Any, None]]]
        ] = None,
    ):
        """添加事件到列表"""
        if event:
            self.__events.append(event)
            return self

    async def execute(self):
        """执行所有事件"""

        for event in self.__events:
            if asyncio.iscoroutinefunction(event):
                await event()
            else:
                event()



from typing import TypeVar

from .status_code import BasicApiId, BasicApiInnerErrorCode, BasicStatusCode


class BasicException(Exception):
    """
    基础异常类
    """

    def __init__(self, code: int, msg: str):
        super().__init__(msg)
        self.code = code
        self.msg = msg


# 定义泛型变量 T，限定为 BasicStatusCode 的子类
T = TypeVar("T", bound=BasicStatusCode)


class BasicCommonException(BasicException):
    def __init__(self, status_code: T, **kwargs):
        super().__init__(code=status_code.code, msg=status_code.msg.format(**kwargs))


# 定义泛型变量 ET，限定为 BasicApiInnerErrorCode 的子类
ET = TypeVar("ET", bound=BasicApiInnerErrorCode)


class BasicApiInnerException(BasicException):
    """
    接口内部异常类的基类
    """

    def __init__(self, status_code: ET, **kwargs):
        super().__init__(code=status_code.code, msg=status_code.msg.format(**kwargs))

    def add_api_id(self, api_id: BasicApiId) -> "BasicApiInnerException":
        self.code = BasicApiInnerErrorCode.gen_standard_code(
            api_id=api_id, inner_error_code=self.code
        )
        return self



from abc import ABC, abstractmethod
from typing import Dict, Optional

from starlette.requests import Request

from algo_backend.schema import AbstractRespVo, BaseRespVo
from algo_backend.utils.utils import gen_random_request_id


class BasicExceptionInterceptor(ABC):
    """
    拦截接口关于schema检验的报错，并返回约定的body
    """

    def __init__(
        self,
        default_vo_type: type(AbstractRespVo) = BaseRespVo,
        url_vo_dict: Optional[Dict[str, type(AbstractRespVo)]] = None,
    ):
        self.default_vo_type: type(AbstractRespVo) = default_vo_type
        # url和响应的匹配策略
        self.url_vo_dict = url_vo_dict or {}

    def get_vo_type(self, url: str) -> type(AbstractRespVo):
        """
        用户可以继承BasicExceptionInterceptor的子类，覆盖这个方法从而实现更加复杂的url和响应体匹配策略
        :param url:
        :return:
        """
        return self.url_vo_dict.get(url, self.default_vo_type)

    @classmethod
    def extract_url(cls, request: Request):
        try:
            url = request.url.path
        except Exception:
            url = ""
        return url

    @classmethod
    def get_request_id(cls, request: Request):
        return request.headers.get("x-request-id", gen_random_request_id())

    @abstractmethod
    def intercept(self, request: Request, exc: Exception): ...



from typing import List

from algo_backend.utils import OsAttrMeta


class LoguruConfig(metaclass=OsAttrMeta):
    LOGGER_PATH: str = "/logger"
    LOG_RETENTION_DAY: int = 60
    DISABLE_LOG_PKG: str = ""
    LOG_ADD_CONTAINED_ID: bool = False
    SAVE_INFO_LEVEL: bool = False
    SAVE_DEBUG_LOG: bool = True

    @classmethod
    def get_disable_log_pkg(cls) -> List[str]:
        if cls.DISABLE_LOG_PKG:
            return cls.DISABLE_LOG_PKG.split(",")
        else:
            return []



import logging
from typing import Optional

from algo_backend.exception import (
    BasicApiId,
    BasicApiInnerException,
    BasicCommonException,
    CommonStatusCode,
)
from algo_backend.schema import AbstractRespVo

logger = logging.getLogger(__name__)


def gen_vo_from_exception(
    vo_cls: type(AbstractRespVo),
    e: Exception,
    api_name: Optional[str] = None,
    request_id: Optional[str] = None,
    api_id: Optional[BasicApiId] = None,
) -> AbstractRespVo:
    """
    将异常转换为vo
    """

    _params = dict(api_name=api_name, request_id=request_id)
    if isinstance(e, BasicCommonException):
        vo = vo_cls.from_exception(e, **_params)
    elif isinstance(e, BasicApiInnerException):
        vo = vo_cls.from_exception(e.add_api_id(api_id=api_id), **_params)
    else:
        vo = vo_cls.from_exception(
            BasicCommonException(CommonStatusCode.UNKNOWN_ERROR, msg=str(e)),
            **_params,
        )

    return vo



import logging

from fastapi.exceptions import StarletteHTTPException
from fastapi.responses import JSONResponse
from starlette.requests import Request

from algo_backend.exception import BasicCommonException, CommonStatusCode

from .common import BasicExceptionInterceptor

logger = logging.getLogger(__name__)


class HTTPExceptionInterceptor(BasicExceptionInterceptor):
    """
    拦截接口关于http异常的报错，并返回约定的body
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def intercept(self, request: Request, exc: StarletteHTTPException):
        url: str = self.extract_url(request)
        reqid: str = self.get_request_id(request)

        logger.info(
            f"req_id: {url} | 错误码: {exc.status_code} | 错误信息: {exc.detail}"
        )
        if exc.status_code == 405:
            e = BasicCommonException(CommonStatusCode.ERROR_REQ_METHOD, url=url)
        elif exc.status_code == 404:
            e = BasicCommonException(CommonStatusCode.ERROR_REQ_URL, url=url)
        else:
            e = BasicCommonException(
                CommonStatusCode.UNKNOWN_ERROR, msg=f"{exc.status_code},{exc.detail}"
            )

        vo = self.get_vo_type(url).from_exception(e, request_id=reqid)

        return JSONResponse(content=vo.model_dump(), status_code=200)



class BasicLogStarter:
    """
    日志启动器，初始化日志格式和落盘等
    """

    def __init__(self, service_name: str):
        """
        :param service_name: 服务名
        """
        self.service_name = service_name

    def setup_log(self):
        """
        执行一些日志设置
        """
        pass



from enum import Enum

from algo_backend.config import ErrorCodeConfig


def add_service_prefix(v: int):
    """
    对于失败的状态码，添加服务号码前缀
    """
    v_str = str(v)
    if len(v_str) == 8:
        return v
    if v is not None and v != 0 and ErrorCodeConfig.SERVICE_PREFIX:
        v = int(f"{ErrorCodeConfig.SERVICE_PREFIX}{v_str}")
    return v


class BasicStatusCode(Enum):
    def __new__(cls, value: int, msg: str):
        assert 0 <= value <= 999999, f"错误码长度不能超过6位，且非负数，{value}"
        obj = object.__new__(cls)
        obj._value_ = obj
        obj._code = value
        obj.msg = msg
        return obj

    @property
    def code(self):
        return add_service_prefix(self._code)

    @property
    def value(self):
        return self.code


class CommonStatusCode(BasicStatusCode):
    SUCCESS = (0, "成功")

    ERROR_REQ_URL = (400001, "请求路径错误")
    ERROR_REQ_METHOD = (400002, "{url}请求方法错误")
    BODY_EMPTY_ERR = (400003, "{url}请求体内容为空")
    PARAM_ERROR = (400006, "接口{url}参数错误,{msg}")
    ERROR_HEADER_NOW_EXISTS = (400007, "接口{url}缺少请求头:{header}")

    UNKNOWN_ERROR = (500000, "未知错误: {msg}")


class BasicApiId(Enum):
    """
    接口编号枚举类
    """

    def __new__(cls, value: int):
        assert 0 <= value <= 999, f"接口编号长度不能超过3位，且非负数，{value}"
        obj = object.__new__(cls)
        obj._value_ = obj
        obj.code = value
        return obj

    @property
    def value(self):
        return self.code

    def bind_api(self):
        """
        装饰器，绑定绑定api id
        """

        def decorator(enum_class):
            enum_class.__api_id__ = self
            return enum_class

        return decorator


class DefaultApiErrorCode(BasicApiId):
    """
    0表示不提供任何ApiId
    """

    DEFAULT_ERROR = 0


class BasicApiInnerErrorCode(Enum):
    def __new__(cls, value: int, msg: str):
        assert 1 <= value <= 99, f"接口内部错误码长度不能超过2位，且为正数，{value}"
        obj = object.__new__(cls)
        obj._value_ = obj
        obj.code = value
        obj.msg = msg
        return obj

    @property
    def value(self):
        return self.code

    @classmethod
    def gen_standard_code(cls, api_id: BasicApiId, inner_error_code: int = 0):
        """
        拼接并生成6位错误码
        """
        api_id = api_id or DefaultApiErrorCode.DEFAULT_ERROR
        return 500000 + api_id.code * 100 + inner_error_code



import os
import socket

from loguru import logger

from algo_backend.config import LoguruConfig as LogConfig

from ..common import BasicLogStarter
from .log_clean import LoguruCleaner
from .log_setup import LoguruSetup


class LoguruStarter(BasicLogStarter):
    """
    容器内日志目录默认是：/logger/服务名
    """

    def __init__(self, service_name: str):
        super().__init__(service_name)

    @property
    def service_log_dir(self):
        return os.path.join(LogConfig.LOGGER_PATH, self.service_name)

    def setup_log(self):
        """
        日志设置
        """

        LoguruSetup.rotate_daily(
            log_dir=self.service_log_dir,
            service_name=self.add_container_id(service_name=self.service_name),
            add_pid_suffix=True,
            save_info=LogConfig.SAVE_INFO_LEVEL,
            save_debug=LogConfig.SAVE_DEBUG_LOG,
        )

        for pkg in LogConfig.get_disable_log_pkg():
            # 忽略一些包的日志
            logger.debug(f"ignore log: {pkg}")
            logger.disable(pkg)

    def run_log_cleaner(self):
        """
        启动定时任务清理日志
        """
        LoguruCleaner.schedule_run(
            log_dir=self.service_log_dir,
            retention_day=LogConfig.LOG_RETENTION_DAY,
        )

    @classmethod
    def add_container_id(cls, service_name: str):
        if not LogConfig.LOG_ADD_CONTAINED_ID:
            logger.info("日志名不增加containerId")
            return service_name

        try:
            socket_hostname = f"-{socket.gethostname()}"
        except:
            socket_hostname = ""
        if service_name in socket_hostname:
            return service_name
        else:
            return f"{service_name}{socket_hostname}"



import asyncio
import logging
from typing import List

from .common import AbstractMetricCollector
from .gc_metrics import PythonGcMetricsCollector, PythonGcObjectMetricsCollector
from .system_metrics import SystemMetricsMonitor

logger = logging.getLogger(__name__)


class MetricsScheduleMonitor:
    """
    定时收集指标
    """

    def __init__(self):
        self.collectors: List[AbstractMetricCollector] = []

    def add(self, collector: AbstractMetricCollector) -> "MetricsScheduleMonitor":
        self.collectors.append(collector)
        return self

    def register_default_collectors(self) -> "MetricsScheduleMonitor":
        """
        注册默认指标收集器
        """
        self.add(SystemMetricsMonitor(interval_sec=30).register_default_collectors())
        self.add(PythonGcMetricsCollector(interval_sec=30))
        self.add(PythonGcObjectMetricsCollector(interval_sec=5))
        return self

    async def run_monitor(self):
        """运行监控任务"""

        if len(self.collectors) == 0:
            self.register_default_collectors()

        logger.info(
            f"Start system metrics monitor, collector num = {len(self.collectors)}"
        )

        async def schedule_collect(collector: AbstractMetricCollector):
            """定时收集指标数据"""
            while True:
                await collector.collect()
                await asyncio.sleep(collector.collect_interval)

        for c in self.collectors:
            asyncio.create_task(schedule_collect(c))



from .common import BasicExceptionInterceptor
from .http import HTTPExceptionInterceptor
from .validate import ValidateExceptionInterceptor

__all__ = [
    "HTTPExceptionInterceptor",
    "ValidateExceptionInterceptor",
    "BasicExceptionInterceptor",
]



import logging

from fastapi.exceptions import RequestValidationError
from fastapi.responses import JSONResponse
from starlette.requests import Request

from algo_backend.exception import BasicCommonException, CommonStatusCode

from .common import BasicExceptionInterceptor

logger = logging.getLogger(__name__)


class ValidateExceptionInterceptor(BasicExceptionInterceptor):
    """
    拦截接口关于schema检验的报错，并返回约定的body
    """

    def __init__(self, exc_msg_len=None, **kwargs):
        super().__init__(**kwargs)
        self.exc_msg_len: int = exc_msg_len or 600

    def intercept(self, request: Request, exc: RequestValidationError) -> JSONResponse:
        """
        拦截接口关于schema检验的报错，并返回约定的body
        """

        reqid: str = self.get_request_id(request)
        url: str = self.extract_url(request)

        logger.error(
            f"Reqid: {reqid} | 请求体参数错误: {str(exc)[: self.exc_msg_len]}..."
        )
        if int(request.headers.get("content-length")) == 0 and request.method in (
            "POST",
            "PUT",
            "PATCH",
        ):
            e = BasicCommonException(
                CommonStatusCode.BODY_EMPTY_ERR,
                url=url,
            )
        else:
            reason: str = self.parse_request_validation_error(exc)
            e = BasicCommonException(
                CommonStatusCode.PARAM_ERROR,
                url=url,
                msg=reason,
            )

        vo = self.get_vo_type(url).from_exception(e, request_id=reqid)

        return JSONResponse(
            status_code=200,  # 或其他合适的 HTTP 状态码
            content=vo.model_dump(),  # 将 Pydantic 模型转换为字典
        )

    @staticmethod
    def parse_request_validation_error(exc: RequestValidationError):
        """
        报错润色
        RequestValidationError([{'type': 'string_type', 'loc': ('body', 'messages', 0, 'role'), 'msg': 'Input should be a valid string...[]},
        {'type': 'string_type', 'loc': ('body', 'messages', 0, 'content'), 'msg': 'Input should be a valid string', 'input': []}])
        转换结果如下：
        '请求参数错误，发现2个字段问题：messages[0].role: Input should be a valid string；messages[0].content: Input should be a valid string'
        """
        error_fields = []
        for error in exc.errors():
            loc = error["loc"]
            # 忽略第一个元素（通常是 body）
            field_path = ".".join(
                f"[{i}]" if isinstance(i, int) else i for i in loc[1:]
            ).replace(".[", "[")
            msg = error["msg"]
            error_fields.append(f"{field_path}: {msg}")
        return f"请求参数错误，发现{len(error_fields)}个字段问题：" + "；".join(
            error_fields
        )



import os
from typing import List, Optional

from prometheus_client import Counter, Histogram
from starlette.middleware.base import BaseHTTPMiddleware
from starlette.requests import Request


class RequestTimeCostMiddleware(BaseHTTPMiddleware):
    def __init__(
        self,
        app,
        buckets: Optional[List[float]] = None,
        ignore_paths: Optional[List[str]] = None,
    ):
        """
        :param buckets: 默认[0.1, 0.5, 1, 3, 5]
        :param ignore_paths: 默认不屏蔽，但是建议用户屏蔽/metrics
        """
        super().__init__(app)

        default_buckets = [0.1, 0.5, 1, 3, 5]
        self.buckets = buckets if buckets is not None else default_buckets
        self.ignore_paths = ignore_paths if ignore_paths is not None else []

        self.request_duration = Histogram(
            "http_request_duration_seconds",
            "Request duration in seconds",
            ["uri", "pid"],
            buckets=self.buckets,
        )

        self.error_requests = Counter(
            "http_error_requests", "error HTTP requests", ["uri", "status"]
        )

    async def dispatch(self, request: Request, call_next):
        # 规范化路径：去掉末尾的斜杠（根路径除外）
        path = request.url.path
        normalized_path = path.rstrip("/") if path != "/" else path

        if normalized_path in self.ignore_paths:
            return await call_next(request)
        else:
            with self.request_duration.labels(
                uri=normalized_path, pid=os.getpid()
            ).time():
                response = await call_next(request)

                # 记录错误请求
                if response.status_code >= 400:
                    self.error_requests.labels(
                        uri=normalized_path, status=response.status_code
                    ).inc()

                return response



from starlette.middleware import Middleware
from starlette.middleware.cors import CORSMiddleware

default_cors_middleware = Middleware(
    CORSMiddleware,
    allow_origins=["*"],
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)



from .error_code_manage import ApiErrorCodeManage, BasicCodeManage
from .exception import BasicApiInnerException, BasicCommonException, BasicException
from .status_code import (
    BasicApiId,
    BasicApiInnerErrorCode,
    BasicStatusCode,
    CommonStatusCode,
    DefaultApiErrorCode,
)

__all__ = [
    "BasicStatusCode",
    "CommonStatusCode",
    "BasicApiId",
    "BasicApiInnerErrorCode",
    "BasicApiInnerException",
    "ApiErrorCodeManage",
    "BasicCodeManage",
    "BasicCommonException",
    "DefaultApiErrorCode",
    "BasicException",
]



import logging

from loguru import logger

from algo_backend.config import LoguruConfig

disable_log_pkg = LoguruConfig.get_disable_log_pkg()


def patch_logging_to_loguru():
    """
    将 Python 原生 logging 系统的所有日志重定向到 loguru
    """

    class LoguruHandler(logging.Handler):
        def emit(self, record):
            # 过滤特定模块的日志
            if any(excluded in record.name for excluded in disable_log_pkg):
                return

            try:
                level = logger.level(record.levelname).name
            except ValueError:
                level = record.levelno

            # 从当前帧开始，找到真正调用 logging 的位置
            frame = logging.currentframe()
            depth = 0  # 从0开始计算，动态确定深度

            # 遍历调用栈，跳过所有 logging 模块的内部调用
            while frame:
                filename = frame.f_code.co_filename
                func_name = frame.f_code.co_name

                # 检查是否为 logging 模块的内部调用
                is_logging_internal = (
                    # 标准库 logging 模块路径
                    "logging" in filename
                    and (
                        filename.endswith("logging/__init__.py")
                        or "/logging/" in filename
                        or "\\logging\\" in filename
                    )
                ) or (
                    # logging 内部函数名
                    func_name
                    in (
                        "callHandlers",
                        "handle",
                        "emit",
                        "handleError",
                        "_log",
                        "makeRecord",
                        "getLogger",
                        "debug",
                        "info",
                        "warning",
                        "error",
                        "exception",
                        "critical",
                    )
                )

                if is_logging_internal:
                    frame = frame.f_back
                    depth += 1
                else:
                    # 找到真实的调用者，跳出循环
                    break

            # 如果找不到合适的帧，回退到默认行为
            if not frame:
                depth = 2

            logger.opt(depth=depth, exception=record.exc_info).log(
                level, record.getMessage()
            )

    # 设置根记录器级别为最低级别，确保所有日志都被处理
    logging.root.setLevel(logging.DEBUG)

    logging.root.handlers = []
    logging.root.addHandler(LoguruHandler())
